from django.contrib import admin
from . import models


class SignalConstraintInline(admin.TabularInline):
    model = models.SignalConstraint
    extra = 0


@admin.register(models.Signal)
class SignalAdmin(admin.ModelAdmin):
    list_display = ('name', 'content_type', 'signal_type', 'active')

    inlines = [SignalConstraintInline]



# Generated by Django 3.2.9 on 2021-12-03 13:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('email_signals', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='signal',
            name='from_email',
            field=models.EmailField(blank=True, help_text='If not set, `settings.EMAIL_SIGNAL_DEFAULT_FROM_EMAIL`             with be used.', max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='signalconstraint',
            name='comparision',
            field=models.CharField(choices=[('exact', 'Is Equal To'), ('iexact', 'Is Equal To (case insensitive)'), ('contains', 'Contains'), ('icontains', 'Contains (case insensitive)'), ('gt', 'Greater Than'), ('gte', 'Greater Than or Equal To'), ('lt', 'Less Than'), ('lte', 'Less Than or Equal To'), ('startswith', 'Starts With'), ('istartswith', 'Starts With (case insensitive)'), ('endswith', 'Ends With'), ('iendswith', 'Ends With (case insensitive)'), ('regex', 'Matches Regular Expression'), ('iregex', 'Matches Regular Expression (case insensitive)'), ('isnull', 'Is Null'), ('isnotnull', 'Is Not Null'), ('istrue', 'Is True'), ('isfalse', 'Is False')], max_length=20),
        ),
        migrations.AlterField(
            model_name='signalconstraint',
            name='signal',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='constraints', to='email_signals.signal'),
        ),
    ]



import typing as _t
from django.db import models
from django.db.models import signals
from django.contrib.contenttypes.models import ContentType
from ckeditor.fields import RichTextField
from . registry import add_to_registry


class EmailSignalMixin:
    """A mixin to add a model to the email_signals app."""

    EMAIL_SIGNAL_MODEL = True

    def email_signal_post_init(self) -> None:
        """Add this model to the registry."""
        add_to_registry(self)

    def email_signal_recipients(self, email_opt: int) -> _t.List[str]:
        """Return a list of email addresses to send the signal to."""
        method_name = f'get_email_signal_emails_{email_opt}'
        if not hasattr(self, method_name):
            raise NotImplementedError(
                f'{self.__class__.__name__} has no method {method_name}'
            )
        return getattr(self, method_name)()


class Signal(models.Model):
    """Stores signals to be raised by the email_signals app."""

    class SignalTypeChoices(models.TextChoices):
        """Choices for the type of signal."""
        pre_save = 'pre_save', 'Pre Save'
        post_save = 'post_save', 'Post Save'
        pre_delete = 'pre_delete', 'Pre Delete'
        post_delete = 'post_delete', 'Post Delete'

    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        verbose_name='Model (Table)'
    )
    plain_text_email = models.TextField(blank=True, null=True)
    html_email = RichTextField(
        blank=True,
        null=True,
        verbose_name='HTML email'
    )
    subject = models.CharField(max_length=255)
    from_email = models.EmailField(
        null=True,
        blank=True,
        help_text='If not set, `settings.EMAIL_SIGNAL_DEFAULT_FROM_EMAIL` \
            with be used.'
    )
    to_emails_opt = models.PositiveIntegerField(
        default=1,
        help_text="The choice of the user to send the email to. For each \
            integer `i`, the method `get_email_signal_emails_<i>` will be \
            called on the model instance. The method should return a list of \
            emails to send to.",
        verbose_name='Mailing list no'
    )
    template = models.CharField(
        max_length=100,
        blank=True,
        null=True,
        help_text="Custom template to use for the email."
    )
    signal_type = models.CharField(
        max_length=20,
        choices=SignalTypeChoices.choices
    )
    active = models.BooleanField(default=True)

    def __str__(self) -> str:
        return f'({self.signal_type}): {self.name}'

    def get_signal_type(self):
        """Return the signal type."""
        return getattr(signals, self.signal_type)

    @ classmethod
    def get_signal_type_from_choice(cls, choice: int):
        """Return the signal type from the choice."""
        return getattr(signals, choice)

    @ classmethod
    def get_choice_from_signal(cls, signal: signals.ModelSignal) -> str:
        """Return the signal type from the signal."""
        if signal == signals.pre_save:
            return cls.SignalTypeChoices.pre_save
        elif signal == signals.post_save:
            return cls.SignalTypeChoices.post_save
        elif signal == signals.pre_delete:
            return cls.SignalTypeChoices.pre_delete
        elif signal == signals.post_delete:
            return cls.SignalTypeChoices.post_delete
        else:
            raise ValueError(f'Unknown signal: {signal}')

    @ classmethod
    def get_for_model_and_signal(
        cls,
        instance: 'Signal',
        signal: signals.ModelSignal
    ) -> models.QuerySet['Signal']:
        """Return the signal for the given model and signal type."""
        return cls.objects.filter(
            content_type=ContentType.objects.get_for_model(instance),
            signal_type=cls.get_choice_from_signal(signal),
            active=True,
        )


class SignalConstraint(models.Model):
    """Stores the constraints for a signal."""

    COMPARISON_CHOICES = (
        ('exact', 'Is Equal To'),
        ('iexact', 'Is Equal To (case insensitive)'),
        ('contains', 'Contains'),
        ('icontains', 'Contains (case insensitive)'),
        ('gt', 'Greater Than'),
        ('gte', 'Greater Than or Equal To'),
        ('lt', 'Less Than'),
        ('lte', 'Less Than or Equal To'),
        ('startswith', 'Starts With'),
        ('istartswith', 'Starts With (case insensitive)'),
        ('endswith', 'Ends With'),
        ('iendswith', 'Ends With (case insensitive)'),
        ('regex', 'Matches Regular Expression'),
        ('iregex', 'Matches Regular Expression (case insensitive)'),
        ('isnull', 'Is Null'),
        ('isnotnull', 'Is Not Null'),
        ('istrue', 'Is True'),
        ('isfalse', 'Is False'),
    )

    signal = models.ForeignKey(
        Signal,
        on_delete=models.CASCADE,
        related_name='constraints'
    )
    param_1 = models.CharField(
        max_length=255,
        verbose_name='Parameter 1',
        help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute.'  # noqa 
    )
    comparison = models.CharField(
        max_length=20,
        choices=COMPARISON_CHOICES
    )
    param_2 = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name='Parameter 2',
        help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute. Also supports primitive values.'

    )

    def __str__(self) -> str:
        return f'{self.signal.name} - {self.comparison} - {self.param_1}'



import typing as _t
import random
import string
from django.db import connection, models
from django.contrib.contenttypes.models import ContentType
from ..models import EmailSignalMixin


def generate_random_string() -> str:
    """Generates a random string."""
    return ''.join(
        random.choice(string.ascii_letters)
        for _ in range(random.randint(3, 10))
    )

def generate_random_email() -> str:
    """Generates a random email."""
    return f'{generate_random_string()}@{generate_random_string()}.com'


class TestCustomerModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a customer model."""
    name = models.CharField(max_length=200, default=generate_random_string)
    email = models.CharField(max_length=200, default=generate_random_email)

    def get_email_signal_emails_1(self) -> _t.List[str]:
        return [self.email]

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label='email_signals',
                model='testcustomermodel',
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute('DROP TABLE IF EXISTS {};'.format(
                    cls._meta.db_table
                ))
            ContentType.objects.filter(
                app_label='email_signals',
                model='testcustomermodel',
            ).delete()

    @classmethod
    def create_record(cls) -> 'TestCustomerModel':
        """Creates a record in the database."""
        rec = cls.objects.create()
        rec.save()
        return rec


class TestCustomerOrderModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a customer order
    model.
    """
    customer = models.ForeignKey(TestCustomerModel, on_delete=models.CASCADE)
    order_number = models.CharField(
        max_length=100,
        default=generate_random_string,
        null=True,
        blank=True
    )

    def get_email_signal_emails_1(self) -> _t.List[str]:
        return [self.customer.email]

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label='email_signals',
                model='testcustomerordermodel',
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute('DROP TABLE IF EXISTS {};'.format(
                    cls._meta.db_table
                ))
            ContentType.objects.filter(
                app_label='email_signals',
                model='testcustomerordermodel',
            ).delete()

    @classmethod
    def create_record(
        cls,
        customer: TestCustomerModel
    ) -> 'TestCustomerOrderModel':
        """Creates a record in the database."""
        rec = cls.objects.create(customer=customer)
        rec.save()
        return rec



import typing as _t
from django.conf import settings
from django.template.loader import render_to_string
from django.core.mail import send_mail as _send_mail


def send_mail(
    subject: str,
    recipient_list: _t.Iterable[str],
    plain_message: _t.Optional[str] = None,
    html_message: _t.Optional[str] = None,
    from_email: _t.Optional[str] = None,
    template: _t.Optional[str] = None,
    context: _t.Optional[_t.Dict[str, _t.Any]] = None,
) -> None:
    """A wrapper for Django's `send_email` function. This will handle cases
    where the `from_email` is not defined and where the user wants to use a
    custom template to render their email.

    Args:
        subject: The subject of the email.
        plain_message: The plain text message of the email.
        html_message: The HTML message of the email.
        recipient_list: The list of recipients of the email.
        from_email: The email address of the sender.
        template: The template to use to render the email.
        context: The context to use to render the email.

    Returns:
        None
    """

    from_email = from_email or settings.EMAIL_SIGNAL_DEFAULT_FROM_EMAIL
    if template:
        html_message = render_to_string(template, context or {})

    _send_mail(
        subject=subject,
        message=plain_message,
        html_message=html_message,
        from_email=from_email,
        recipient_list=recipient_list,
    )



from django.apps import AppConfig, apps


class EmailSignalsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'email_signals'
    verbose_name = 'Email Signals'

    def ready(self):
        """Find models which should be registered by this app and register
        them.
        """
        from .registry import add_to_registry
        for model in apps.get_models():
            if getattr(model, 'EMAIL_SIGNAL_MODEL', False):
                add_to_registry(model)

        # Order matters! We need to register models before we can import
        # `signals` models as it depends on registered models.
        from . import signals
        signals.setup()



"""Contains collection of models that have been registered to have signals
dispatched for. Also contains helper functions relating to the registry.
"""

import typing as _t
from django.db.models import Model


registered_models: _t.Dict[str, Model] = {}


def model_str(model: Model) -> str:
    """Returns a string representation of the model which can be used to
    identify both the app label and the model name.

    Args:
        model: The model to get the string representation of.

    Returns:
        A string representation of the model.
    """
    return f"{model._meta.app_label}.{model._meta.model_name}"


def add_to_registry(model: Model) -> None:
    """Adds a model to the registry.

    Args:
        model: The model to add to the registry.
    """
    registered_models[model_str(model)] = model


def model_in_registry(model: Model) -> bool:
    """Checks if a model is in the registry.

    Args:
        model: The model to check.

    Returns:
        True if the model is in the registry, False otherwise.
    """
    return model_str(model) in registered_models



from types import SimpleNamespace
from django.test import SimpleTestCase
from .testcase import EmailSignalTestCase
from .. import utils
from ..models import SignalConstraint


class TestConvertToPrimitive(SimpleTestCase):
    """Unittests for the `convert_to_primitive` utility function."""

    def test_convert_to_true(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'true'
        """
        self.assertEqual(utils.convert_to_primitive('true'), True)
        self.assertEqual(utils.convert_to_primitive('tRuE'), True)

    def test_convert_to_false(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'false'
        """
        self.assertEqual(utils.convert_to_primitive('false'), False)
        self.assertEqual(utils.convert_to_primitive('fAlSe'), False)

    def test_convert_to_none(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'none'
        """
        self.assertEqual(utils.convert_to_primitive('none'), None)
        self.assertEqual(utils.convert_to_primitive('nOnE'), None)
        self.assertEqual(utils.convert_to_primitive('null'), None)
        self.assertEqual(utils.convert_to_primitive('nuLL'), None)

    def test_convert_to_float(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is a float
        """
        self.assertEqual(utils.convert_to_primitive('1.0'), 1.0)
        self.assertEqual(utils.convert_to_primitive('1.1'), 1.1)

    def test_convert_to_bad_float(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is not a flat but has a '.' in it. It should return the value back as
        a string.
        """
        self.assertEqual(utils.convert_to_primitive('1.1.1'), '1.1.1')

    def test_convert_to_int(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is an int
        """
        self.assertEqual(utils.convert_to_primitive('1'), 1)
        self.assertEqual(utils.convert_to_primitive('2'), 2)

    def test_failed_convert_to_primitive(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is not a valid value.
        """
        self.assertEqual(utils.convert_to_primitive('a'), 'a')


class TestGetParamFromObj(EmailSignalTestCase):
    """Unittests for the `get_param_from_obj` utility function."""

    def test_dict(self):
        """Test the function returns the current value from a `dict`."""
        self.assertEqual(utils.get_param_from_obj('a', {'a': 1}), (True, 1))

    def test_multi_layered_dict(self):
        """Test the function returns the current value from a `dict`."""
        self.assertEqual(
            utils.get_param_from_obj('a.b.c', {'a': {'b': {'c': 1}}}),
            (True, 1),
        )

    def test_iterable(self):
        """Test the function returns the current value from a iterable."""
        self.assertEqual(utils.get_param_from_obj('0', [1, 2]), (True, 1))
        self.assertEqual(utils.get_param_from_obj('0', (1, 2)), (True, 1))

    def test_multi_layered_iterable(self):
        """Test the function returns the current value from a iterable."""
        self.assertEqual(
            utils.get_param_from_obj(
                '0.1',
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            (True, 2),
        )
        self.assertEqual(
            utils.get_param_from_obj(
                '0.1',
                ((1, 2, 3), (4, 5, 6), (7, 8, 9))
            ),
            (True, 2),
        )

    def test_simplenamespace(self):
        """Test the function returns the current value from a `SimpleNamespace`
        object.
        """
        self.assertEqual(
            utils.get_param_from_obj('a', SimpleNamespace(a=1)),
            (True, 1),
        )

    def test_model_instance(self):
        """Test that the function is able to travel into a model instance."""

        self.Customer.objects.create(name='Util Test Customer').save()

        self.assertEqual(
            utils.get_param_from_obj(
                '0.name',
                self.Customer.objects.filter(name='Util Test Customer')
            ),
            (True, 'Util Test Customer'),
        )
